﻿'*** References ***
'Lars Kobbe describes a way to get Midi data from an aif file generated by Apple Garageband
'exporting a sample as a loop and/or into the loop library and save it as aif.
'After this, the Midi data can be extracted into a file with a binary (hex) editor or by his AppleScript
'http://www.larskobbe.de/midi-export-in-apples-garageband/
'The description is either inaccurate or specific to Apple-Midi-Files, so not assignable to Windows.
'Nice work anyway

'The description "Creating AIFF Audio Formatted files" by Paul Bourke
'has been helpful to walk the chunks and not having to compare byte-by-byte whatever is coming
'up to the right chunk id. http://www.paulbourke.net/dataformats/audio/
'(Also appreciate his remarkable work one directory level above this link)

Module Module1
	'This should be in ressources but it has to be written here anyway for you to read
	Dim UsageText() =
	{vbNewLine & _
	 "MidiFromAIF extracts the first Midi chunk from an aif(f) soundfile" & vbNewLine & _
	 "and saves it to a mid-file of the same name in the same directory." & vbNewLine & _
	 "Freeware/Open Source: Everybody may freely use or copy the program." & vbNewLine & _
	 "Source: midifromaif.codeplex.com - Author: Dieter Bäumler" & vbNewLine & vbNewLine & _
	 "Usage:" & vbNewLine & _
	 "Either drag one or more aif(f) files onto the program/icon or" & vbNewLine & _
	 "call at the command line: midifromaif <aif(f) file> [<aif(f) file>] [...]" & vbNewLine & _
	 "File names containing spaces or comma need to be ""enclosed"" this way." & vbNewLine,
		vbNewLine & _
	 "MidiFromAIF extrahiert den ersten Midi-Chunk aus einer Aif(f)-Datei" & vbNewLine & _
	 "und speichert ihn als Mid-Datei desselben Namens im selben Verzeichnis." & vbNewLine & _
	 "Freeware/Open Source: Jeder darf das Programm nutzen und kopieren." & vbNewLine & _
	 "Quelle: midifromaif.codeplex.com - Autor: Dieter Bäumler" & vbNewLine & vbNewLine & _
	 "Anwendung:" & vbNewLine & _
	 "Entweder mit der Maus Aif(f)-Dateien auf das Programmsymbol ziehen oder" & vbNewLine & _
	 "Kommandozeilen-Aufruf: midifromaif <Aif(f)-Datei> [<Aif(f)-Datei>] [...]" & vbNewLine & _
	 "Dateinamen, die Leerzeichen oder Komma enthalten, sind so ""einzufassen""." & vbNewLine
	}

	Dim pause As Boolean = False 'flag to pause before window possibly closes

	'There are only two languages by now, so the values 0/1 for indexing the above can be set this way
	Dim currentLanguage = -(System.Globalization.CultureInfo.CurrentCulture.TwoLetterISOLanguageName = "de")

	'current OS uses Little-Endian byte order to save multibyte values (or unlikely does it reverse)
	Dim littleEndian As Boolean = BitConverter.IsLittleEndian

	Sub Main()

		'set language for debugging purpose (uncomment both lines)
		'My.Application.ChangeUICulture("fr-FR")	 'french just love to be spoken to in English ;-)
		'currentLanguage = 0

		'get command line parameter string
		Dim CommandStr As String = Command().Trim

		'if empty or help is wanted then give some and end program
		If CommandStr = "" Or "#/h#/help#-h#-help#--help#".Contains("#" & CommandStr.ToLower & "#") Then Usage()

		'compile regular expression for extracting filenames in ""
		Dim rx As New System.Text.RegularExpressions.Regex(""".[^""]+""")
		'compile regular expression for splitting filenames on one or more spaces or ","
		Dim splt As New System.Text.RegularExpressions.Regex("[ ,]+")

		'extract filenames enclosed by ""
		Dim FileList As List(Of String) = rx.Matches(CommandStr).OfType(Of System.Text.RegularExpressions.Match).
			SelectMany(Of String)(Function(it As System.Text.RegularExpressions.Match) {it.Value.Trim(Chr(34))}).ToList
		'into a new list

		'add those filenames to the list just separated by space or "," from the remaining command line
		FileList.AddRange(splt.Split(rx.Replace(CommandStr, ""), 1000))

		'clean list from double and empty entries to parameter array 
		Dim params() As String = FileList.Where(Function(f) f.Length).Distinct.ToArray
		If params.Length = 0 Then Usage() 'if there is nothing to do, show Usage and end

		For Each param As String In params 'handle each parameter given
			'handle the parameter as a filename pattern with wildcards, so cut it from the path
			Dim fpattern As String = param.Split({System.IO.Path.DirectorySeparatorChar, System.IO.Path.AltDirectorySeparatorChar, System.IO.Path.VolumeSeparatorChar}).Last
			Dim fdir As String
			'form directory path if there is one or don't
			If fpattern = param Then fdir = "." Else fdir = Left(param, param.Length - fpattern.Length)
			'get files in directory on a possible pattern 
			Dim thisFiles() As String = My.Computer.FileSystem.GetFiles(fdir, FileIO.SearchOption.SearchTopLevelOnly, fpattern).ToArray
			'if there is nothing there will be an error, ...
			If thisFiles.Length Then
				For Each fpath As String In thisFiles	'... otherwise:
					Dim br As System.IO.BinaryReader 'run through current input file
					Try
						'initiate reader to file (delivering ASCII on reading text)
						br = New System.IO.BinaryReader(New System.IO.FileStream(fpath, IO.FileMode.Open, IO.FileAccess.Read), System.Text.Encoding.ASCII)
					Catch ex As Exception
						Console.WriteLine(fpath & ": " & My.Resources.Err & " (" & vbNewLine & ex.Message & ")" & vbNewLine) : pause = True
						Continue For
					End Try

					'aif(f) files start with "FORM", so if this doesn't it isn't one
					If br.ReadChars(4) <> "FORM" Then
						Console.WriteLine(fpath & " " & My.Resources.noAIF & vbNewLine)
						pause = True
						Continue For
					End If

					'master chunk size follows (all chunk sizes are to be upvalued to even)
					Dim allsize As UInt32 = readUInt32(br)
					allsize += 8 + (allsize And 1)
					'we take this value (master chunksitze + length bytes + 4 master chunk ID chars + possible parity correction) to preview EOF

					'aif(f)-files start with chunk ID AIFF, if this file doesn't, it isn't one
					If br.ReadChars(4) <> "AIFF" Then
						Console.WriteLine(fpath & " " & My.Resources.noAIF & vbNewLine)
						pause = True
						Continue For
					End If

					'now walk the sub chunks

					Dim chnk As String = ""	'holds chunk ID

					'before EOF and if not a Midi chunk has been processed lately
					While (br.BaseStream.Position < allsize) And chnk <> ".mid"
						chnk = br.ReadChars(4) ' read chunk ID
						Dim chkl As UInt32 = readUInt32(br)	'get the chunk length in right byte order

						'chunk information for debug reasons, uncomment if needed
						'Console.WriteLine(vbTab & vbTab & chnk & vbTab & chkl)

						If chnk = ".mid" Then	'if we got a Midi chunk
							Dim buffer() As Byte = br.ReadBytes(chkl)	'fill buffer with Midi chunk data

							'construct output filename

							Dim tfname As String
							Dim fext As String = System.IO.Path.GetExtension(fpath)	'get file extension 
							If fext = ".aiff" Or fext = ".aif" Then	'if it is aif(f) 
								tfname = System.IO.Path.GetFileNameWithoutExtension(fpath) 'remove it (it'll get ".mid")
							Else
								tfname = fpath 'else don't remove extension, ".mid" will be appended
							End If

							Dim fnctr As Integer = 0 'midi file gets this number if filename already exists
							Dim fname As String

							Do 'count up number as long as the construced filename exists
								fname = tfname & IIf(fnctr = 0, "", " (" & fnctr & ")") & ".mid"
								fnctr += 1
							Loop While FileIO.FileSystem.FileExists(fname)

							Try	'save file
								My.Computer.FileSystem.WriteAllBytes(fname, buffer, True)
							Catch ex As Exception
								Console.WriteLine(fname & ": " & My.Resources.Err & " (" & ex.Message & ")")
								pause = True
							End Try

							Console.WriteLine(My.Resources.extracted & ": " & fname & vbNewLine)
							Exit While 'don't look for Midi chunks in this file anymore
						Else 'in case it hasn't been a Midi chunk
							br.BaseStream.Position += chkl + (chkl And 1)	'skip chunk length bytes (+parity correction)
						End If

					End While	'next chunk

					'all chunks in file handled up to a possible Midi chunk
					If chnk <> ".mid" Then Console.WriteLine(fpath & ": " & My.Resources.noMIDI & vbNewLine) : pause = True

				Next 'next file of parameter pattern
			Else 'pattern didn't deliver any files
				Console.WriteLine(param & ": " & My.Resources.FileNotFound)
				pause = True
			End If
		Next

		If pause Then	'something happened to cause a message which has to be seen by the user before window possibly closes
			Console.WriteLine(vbNewLine & My.Resources.PressAnyKey)
			Console.ReadKey(False) 'so wait for the [Any]-Key
		End If
	End Sub

	'gets unsigned integer value in the right byte order saved reverse in the file or say "unreversed"
	Function readUInt32(binreader As System.IO.BinaryReader) As UInt32
		If littleEndian Then 'if systems byte order is (most likely) this
			'reverse 4 read bytes for the value to return
			Return BitConverter.ToUInt32(binreader.ReadBytes(4).Reverse.ToArray, 0)
		Else 'read and return normally (won't happen, but who knows the system this is running on some day)
			Return binreader.ReadUInt32
		End If
	End Function

	Sub Usage()	'throw usage text in proper language
		Console.WriteLine(UsageText(currentLanguage))
		End
	End Sub

End Module
